from .core import ActionProxy
from .context import current_context
from ..expression import eval_expr
import inspect


action_decorators = []


def action_decorator(name):
    """Decorator to register an action decorator
    """
    def decorator(cls):
        action_decorators.append((name, cls))
        return cls
    return decorator


@action_decorator("with_items")
class WithItemsDecorator(ActionProxy):
    """Executes the proxyied action in a loop
    """
    def __init__(self, action, items):
        self.action = action
        self.items = items

    def execute(self):
        """Executes the action using a context where vars will be the kwargs
        """
        items = eval_expr(self.items, current_context.vars)
        if not items:
            return []
        for item in items:
            opts = eval_expr(self.action.options, dict(current_context.vars, item=item))
            rv = self.action._execute(opts)
            if inspect.isgenerator(rv):
                # see Action class for why this can be a generator
                try:
                    rv.next() # actual return value of the function, ignored
                    rv.next() # end function execution
                except StopIteration:
                    pass


@action_decorator("when")
class WhenDecorator(ActionProxy):
    """Executes the proxyies action only if an expression evals to True
    """
    def __init__(self, action, test):
        self.action = action
        self.test = test

    def is_valid(self, context):
        if self.test is None:
            valid = True
        else:
            valid = eval_expr(self.test, context.vars)
        if valid:
            return self.action.is_valid(context)
        return False


@action_decorator("ignore_errors")
class IgnoreErrorDecorator(ActionProxy):
    """Catches all errors raised by the action and ignores them
    """
    def __init__(self, action, enabled=True):
        self.action = action
        self.enabled = enabled

    def execute(self, *args, **kwargs):
        try:
            return self.action.execute(*args, **kwargs)
        except:
            if not self.enabled:
                raise


@action_decorator("only_once")
class OnlyOnceDecorator(ActionProxy):
    """Only executes the proxyied action if it was not already executed
    in the current context
    """
    def __init__(self, action, enabled=True):
        self.action = action
        self.enabled = enabled

    def is_valid(self, context):
        return not self.enabled or self.action.name not in current_context.executed_actions


with_items = WithItemsDecorator
when = WhenDecorator
ignore_errors = IgnoreErrorDecorator
only_once = OnlyOnceDecorator


from werkzeug.utils import import_string as wz_import_string
from werkzeug.local import LocalProxy
from flask import (url_for, Markup, json, request, _request_ctx_stack,\
                   has_request_context)
import imp
import functools
import re
from slugify import slugify
import logging
import yaml
import speaklater
import os
from contextlib import contextmanager


logger = logging.getLogger("frasco")
# logger.setLevel(logging.DEBUG)
# logger.addHandler(logging.StreamHandler())


class RequirementMissingError(Exception):
    pass


class ClassBoundDecoratorMixin(object):
    def _wrap_func(self, func):
        @functools.wraps(func)
        def bound_func(*args, **kwargs):
            return self._call_func(func, *args, **kwargs)
        bound_func.unbound_func = func
        return bound_func

    def _call_func(self, func, *args, **kwargs):
        if not hasattr(self, "obj") or self.obj is None:
            return func(*args, **kwargs)
        return func(self.obj, *args, **kwargs)

    def __get__(self, obj, cls):
        self.obj = obj
        return self


def import_string(impstr, attr=None):
    """Imports a string. Can import an attribute of the imported
    class/module using a double colon as a separator
    """
    if "::" in impstr:
        impstr, attr = impstr.split("::")
    imported = wz_import_string(impstr)
    if attr is not None:
        return getobjpath(imported, attr)
    return imported


def getobjpath(obj, path):
    """Returns an item or attribute of the object recursively.
    Item names are specified between brackets, eg: [item].
    Attribute names are prefixed with a dot (the first one is optional), eg: .attr
    Example: getobjpath(obj, "attr1.attr2[item].attr3")
    """
    if not path:
        return obj
    if path.startswith("["):
        item = path[1:path.index("]")]
        return getobjpath(obj[item], path[len(item) + 2:])
    if path.startswith("."):
        path = path[1:]
    if "." in path or "[" in path:
        dot_idx = path.find(".")
        bracket_idx = path.find("[")
        if dot_idx == -1 or bracket_idx < dot_idx:
            idx = bracket_idx
            next_idx = idx
        else:
            idx = dot_idx
            next_idx = idx + 1
        attr = path[:idx]
        return getobjpath(getattr(obj, attr), path[next_idx:])
    return getattr(obj, path)


def find_classes_in_module(module, clstypes):
    """Find classes of clstypes in module
    """
    classes = []
    for item in dir(module):
        item = getattr(module, item)
        try:
            for cls in clstypes:
                if issubclass(item, cls) and item != cls:
                    classes.append(item)
        except Exception as e:
            pass
    return classes


def remove_yaml_frontmatter(source, return_frontmatter=False):
    """If there's one, remove the YAML front-matter from the source
    """
    if source.startswith("---\n"):
        frontmatter_end = source.find("\n---\n", 4)
        if frontmatter_end == -1:
            frontmatter = source
            source = ""
        else:
            frontmatter = source[0:frontmatter_end]
            source = source[frontmatter_end + 5:]
        if return_frontmatter:
            return (source, frontmatter)
        return source
    if return_frontmatter:
        return (source, None)
    return source


def parse_yaml_frontmatter(source):
    source, frontmatter = remove_yaml_frontmatter(source, True)
    if frontmatter:
        return (yaml.load(frontmatter), source)
    return (None, source)


def populate_obj(obj, attrs):
    """Populates an object's attributes using the provided dict
    """
    for k, v in attrs.iteritems():
        setattr(obj, k, v)


def copy_extra_feature_options(feature, target, prefix=""):
    for k, v in feature.options.iteritems():
        if not feature.defaults or k not in feature.defaults:
            target["%s%s" % (prefix, k.upper())] = v


def url_for_static(filename, **kwargs):
    """Shortcut function for url_for('static', filename=filename)
    """
    return url_for('static', filename=filename, **kwargs)


def url_for_same(**overrides):
    return url_for(request.endpoint, **dict(dict(request.args,
        **request.view_args), **overrides))


def wrap_in_markup(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        return Markup(f(*args, **kwargs))
    return wrapper


def deep_update_dict(a, b):
    for k, v in b.iteritems():
        if k not in a:
            a[k] = v
        elif isinstance(a[k], dict) and isinstance(v, dict):
            deep_update_dict(a[k], v)
        elif isinstance(a[k], list) and isinstance(v, list):
            a[k].extend(v)
        elif isinstance(v, list) and not isinstance(a[k], list):
            a[k] = [a[k]] + v
        else:
            a[k] = v


def make_kwarg_validator(name, validator_func):
    if not isinstance(name, tuple):
        name = (name,)
    def decorator_gen(**kwargs):
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kw):
                values = {n: kw.get(n) for n in name}
                if not validator_func(**dict(kwargs, **values)):
                    abort(400)
                return func(*args, **kw)
            return wrapper
        return decorator
    return decorator_gen


def kwarg_validator(name):
    def decorator(validator_func):
        return make_kwarg_validator(name, validator_func)
    return decorator


class UnknownValue(object):
    pass

unknown_value = UnknownValue()


class DictObject(object):
    def __init__(self, dct):
        for k, v in dct.iteritems():
            setattr(self, k, v)


class AttrDict(dict):
    """Dict which keys are accessible as attributes
    """
    def __getattr__(self, name):
        return self[name]

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        del self[name]

    def get_or_raise(self, name, message):
        try:
            return self[name]
        except KeyError:
            raise KeyError(message)


class JSONEncoder(json.JSONEncoder):
    """A JSONEncoder which always activates the for_json feature
    """
    def __init__(self, *args, **kwargs):
        kwargs["for_json"] = True
        super(JSONEncoder, self).__init__(*args, **kwargs)

    def default(self, o):
        if isinstance(o, speaklater._LazyString):
            return o.value
        if isinstance(o, set):
            return list(o)
        return json.JSONEncoder.default(self, o)


reloader_extra_dirs = ["actions", "features", "views", "services"]


def get_reloader_extra_files(root='.'):
    extra_files = []
    for filename in os.listdir(root):
        pathname = os.path.join(root, filename)
        if os.path.isdir(pathname) and filename in reloader_extra_dirs:
            for path, _, filenames in os.walk(pathname):
                for f in filenames:
                    if f.endswith('.html') or f.endswith('.yml'):
                        extra_files.append(os.path.join(path, f))
        elif filename.endswith('.yml'):
            extra_files.append(pathname)
    return extra_files


class ContextStack(object):
    def __init__(self, top=None, default_item=None, allow_nested=True, ignore_nested=False):
        self.stack = []
        self.top = top
        self.default_item = default_item
        self.allow_nested = allow_nested
        self.ignore_nested = ignore_nested

    def push(self, item=unknown_value):
        if self.stack and not self.allow_nested:
            raise RuntimeError('Context does not support nesting')
        if self.stack and self.ignore_nested:
            item = self.top
        elif item is unknown_value:
            if callable(self.default_item):
                item = self.default_item()
            else:
                item = self.default_item
        self.stack.append(self.top)
        self.top = item
        return item

    def pop(self):
        top = self.top
        self.top = self.stack.pop()
        return top

    @contextmanager
    def ctx(self, item=unknown_value):
        item = self.push(item)
        try:
            yield item
        finally:
            self.pop()

    def __call__(self, *args, **kwargs):
        return self.ctx(*args, **kwargs)

    def make_proxy(self):
        return LocalProxy(lambda: self.top)

    def make_context_mixin(self):
        class ContextMixin(object):
            def __enter__(s):
                self.push(s)
                return s
            def __exit__(s, exc_type, exc_val, exc_tb):
                self.pop()
                return False
        return ContextMixin


def context_stack_on_request_context(name, cls=ContextStack):
    def _get_object():
        if has_request_context() and not hasattr(_request_ctx_stack.top, name):
            setattr(_request_ctx_stack.top, name, cls())
        return getattr(_request_ctx_stack.top, name, None)
    return LocalProxy(_get_object)


class DelayedCallsContext(ContextStack):
    def __init__(self):
        super(DelayedCallsContext, self).__init__(default_item=list, ignore_nested=True)

    def call(self, func, args, kwargs):
        if self.top is not None:
            self.top.append((func, args, kwargs))
            return False
        else:
            func(*args, **kwargs)
            return True

    def pop(self, drop_calls=False):
        top = super(DelayedCallsContext, self).pop()
        if not drop_calls and not self.stack:
            for func, args, kwargs in top:
                func(*args, **kwargs)

    def proxy(self, func):
        @functools.wraps(func)
        def proxy(*args, **kwargs):
            return self.call(func, args, kwargs)
        return proxy


from .helpers import ReturnValueException
from .context import current_context
from .registry import ActionRegistry
from ..utils import populate_obj, slugify
from ..signals import signal
import flask


common_actions = ActionRegistry()



@common_actions.action()
def exit_context(trigger_action_group=None):
    """Triggers a ContextExitException
    """
    current_context.exit(trigger_action_group=trigger_action_group)


@common_actions.action(default_option="group")
def trigger_group(group):
    """Triggers a TriggerActionGroupException
    """
    current_context.trigger_action_group(group)


@common_actions.action("return", default_option="value")
def return_value(value=None):
    raise ReturnValueException(value)


@common_actions.action(default_option="vars")
def export_var(vars):
    if not isinstance(vars, list):
        vars = [vars]
    current_context.export_vars.extend(vars)


@common_actions.action('set')
def set_var(**vars):
    """Sets context variables using the key/value provided in the options
    """
    current_context.vars.update(**vars)


@common_actions.action()
def incr(**vars):
    """Increments context variables
    """
    for k, v in vars:
        current_context.vars.setdefault(k, 0)
        current_context[k] += v


@common_actions.action('set_default')
def set_default_var(**vars):
    """Sets context variables using the key/value provided in the options
    """
    for k, v in vars.iteritems():
        current_context.vars.setdefault(k, v)


@common_actions.action()
def append_to(obj, items=None):
    obj.extend(items)


@common_actions.action()
def remove_from(obj, items=None):
    for i in items:
        obj.remove(i)


@common_actions.action()
def update_dict(obj, **items):
    obj.update(items)


@common_actions.action()
def pop_dict(obj, keys):
    for key in keys:
        obj.pop(key, None)


@common_actions.action()
def append_to_attr(obj, **attrs):
    for name, value in attrs.iteritems():
        if not hasattr(obj, name) or getattr(obj, name) is None:
            setattr(obj, name, [])
        getattr(obj, name).append(value)


@common_actions.action()
def remove_from_attr(obj, **attrs):
    for name, value in attrs.iteritems():
        if hasattr(obj, name):
            getattr(obj, name).remove(value)


@common_actions.action()
def update_in_attr(obj, **attrs):
    for name, value in attrs.iteritems():
        if not hasattr(obj, name) or getattr(obj, name) is None:
            setattr(obj, name, {})
        getattr(obj, name).update(value)


@common_actions.action()
def pop_in_attr(obj, **attrs):
    for name, value in attrs.iteritems():
        if not hasattr(obj, name) or getattr(obj, name) is None:
            setattr(obj, name, {})
        getattr(obj, name).pop(value, None)


@common_actions.action()
def populate_obj(obj, **attrs):
    populate_obj(obj, attrs)


@common_actions.action()
def incr_obj(obj, **attrs):
    """Increments context variables
    """
    for name, value in attrs.iteritems():
        v = getattr(obj, name, None)
        if not hasattr(obj, name) or v is None:
            v = 0
        setattr(obj, name, v + value)


@common_actions.action()
def call(obj, func_name, **kwargs):
    return getattr(obj, func_name)(**kwargs)


@common_actions.action('do', default_option='code')
def exec_python(code):
    return eval(code, globals(), current_context.vars)


@common_actions.action(default_option="code")
def abort(code, **kwargs):
    """Triggers flask.abort() using the specified http code
    """
    flask.abort(code, **kwargs)


@common_actions.action(default_option="view")
def redirect(view=None, url=None, **kwargs):
    """Redirects to the specified view or url
    """
    if view:
        if url:
            kwargs["url"] = url
        url = flask.url_for(view, **kwargs)
    current_context.exit(flask.redirect(url))


@common_actions.action(default_option="msg")
def flash(msg, label="success"):
    """Adds a flash message
    """
    flask.flash(msg, label)


@common_actions.action(default_option="__obj")
def jsonify(__obj=None, **obj):
    if __obj:
        obj = __obj
    current_context.exit(flask.jsonify(**obj))


@common_actions.action(default_option="filename")
def render_template(filename, **kwargs):
    return flask.render_template(filename, **kwargs)


@common_actions.action()
def set_headers(**kwargs):
    current_context.vars.response.headers.extend(kwargs)


@common_actions.action(default_option="msg")
def log(msg, level="info"):
    getattr(flask.current_app.logger, level)(msg)


@common_actions.action(default_option="msg")
def log_debug(msg):
    flask.current_app.logger.debug(msg)


@common_actions.action(default_option="name")
def send_signal(name, **kwargs):
    s = signal(name)
    s.send(**kwargs)


@common_actions.action(default_option="text", as_="slug")
def slugify(text, **kwargs):
    return slugify(text, **kwargs)


from frasco.testsuite import FrascoTestCase
from frasco.expression import Parser, Expression
from frasco.utils import AttrDict


class ExpressionTestCase(FrascoTestCase):
    def assert_nodes(self, results, nodes):
        if len(nodes) != len(results):
            raise AssertionError("Nodes lists are not of the same length")
        for i in xrange(0, len(nodes)):
            if nodes[i] != repr(results[i]):
                raise AssertionError("Node do not match, got '%s', expected '%s'" % (repr(results[i]), nodes[i]))

    def test_parser(self):
        p = Parser()
        self.assert_nodes(p.parse("hello world"), ["DataNode(hello world)"])
        self.assert_nodes(p.parse("$foo"), ["VarNode(foo)"])
        self.assert_nodes(p.parse("$foo str"), ["VarNode(foo)", "DataNode( str)"])
        self.assert_nodes(p.parse("\$foo str"), ["DataNode($foo str)"])
        self.assert_nodes(p.parse("hello${foo}world"), ["DataNode(hello)", "VarNode(foo)", "DataNode(world)"])
        self.assert_nodes(p.parse("$foo.bar"), ["GetAttrNode(VarNode(foo), bar)"])
        self.assert_nodes(p.parse("$foo[bar]"), ["GetItemNode(VarNode(foo), bar)"])
        self.assert_nodes(p.parse("$foo.bar[baz]"), ["GetItemNode(GetAttrNode(VarNode(foo), bar), baz)"])
        self.assert_nodes(p.parse("$foo[bar].baz"), ["GetAttrNode(GetItemNode(VarNode(foo), bar), baz)"])
        self.assert_nodes(p.parse("!$foo"), ["NotNode(VarNode(foo))"])
        self.assert_nodes(p.parse("$foo|20"), ["DefaultNode(VarNode(foo), '20')"])
        self.assert_nodes(p.parse("$foo|$bar"), ["DefaultNode(VarNode(foo), VarNode(bar))"])
        self.assert_nodes(p.parse("$foo && $bar"), ["AndNode(VarNode(foo), VarNode(bar))"])
        self.assert_nodes(p.parse("$foo || $bar"), ["OrNode(VarNode(foo), VarNode(bar))"])
        self.assert_nodes(p.parse("$foo || $bar && $baz"), ["OrNode(VarNode(foo), AndNode(VarNode(bar), VarNode(baz)))"])
        self.assert_nodes(p.parse("$foo && !$bar"), ["AndNode(VarNode(foo), NotNode(VarNode(bar)))"])
        self.assert_nodes(p.parse("$foo && $bar|20"), ["AndNode(VarNode(foo), DefaultNode(VarNode(bar), '20'))"])
        self.assert_nodes(p.parse("{{ 2 + 2 }}"), ["EvalNode(2 + 2)"])

    def test_compile(self):
        p = Parser()
        self.assert_isinstance(p.compile("hello world"), str)
        self.assert_isinstance(p.compile("hello $name"), Expression)

    def test_eval_expression(self):
        p = Parser()
        c = {"foo": "hello", "t": True, "f": False, "data": AttrDict({"a": "aa", "b": {"c": "cc"}})}
        self.assert_equal(p.compile("$foo").eval(c), "hello")
        self.assert_equal(p.compile("$data.a").eval(c), "aa")
        self.assert_equal(p.compile("$data[a]").eval(c), "aa")
        self.assert_equal(p.compile("$data.b[c]").eval(c), "cc")
        self.assert_false(p.compile("!$t").eval(c))
        self.assert_false(p.compile("$t && $f").eval(c))
        self.assert_true(p.compile("$t || $f").eval(c))
        self.assert_equal(p.compile("{{ 2 + 2 }}").eval(c), 4)
        self.assert_equal(p.compile("$undefined|hi").eval(c), "hi")


from flask.cli import DispatchingApp, pass_script_info
import click
import os
import sys
import pipes
from collections import OrderedDict
import honcho.environ
import honcho.manager
import honcho.process
from ..utils import get_reloader_extra_files
from ..commands import shell_exec


def join_process_args(args):
    return " ".join([(pipes.quote(s) if not s.startswith("$") else s) for s in args])


def procfile_processes(processes, info):
    procfile = OrderedDict()
    for name, args in processes:
        args = list(args)
        base = []
        if args[0] == "frasco":
            del args[0]
            base = ["python", "-m", "frasco"]
            if info.debug:
                base.append("--debug")
            if info.app_import_path:
                base.extend(["-a", info.app_import_path])
        procfile[name] = join_process_args(base + map(str, args))
    return procfile


def get_app_processes(info):
    app = info.load_app()
    serve_process = ("web", ["frasco", "serve", "--host", "$HOST", "--port", "$PORT"])
    return procfile_processes([serve_process] + app.processes, info)


def start_honcho(processes, env=None):
    manager = honcho.manager.Manager()
    for p in honcho.environ.expand_processes(processes, env=env):
        e = os.environ.copy()
        e.update(p.env)
        manager.add_process(p.name, p.cmd, env=e)
    manager.loop()
    sys.exit(manager.returncode)


@click.command('run', short_help='Runs a development server and associated workers.')
@click.option('--host', '-h', default='127.0.0.1',
              help='The interface to bind to.')
@click.option('--port', '-p', default=5000,
              help='The port to bind to.')
@click.option('--procfile/--no-procfile', default=False,
              help='Whether to look for a Procfile.')
@click.argument('process', required=False)
@pass_script_info
def run_command(info, host, port, procfile, process):
    processes = None
    if procfile:
        for filename in ["Procfile.%s" % info.env, "Procfile"]:
            if os.path.exists(filename):
                click.echo("Using %s" % filename)
                with open(filename) as f:
                    content = f.read()
                processes = honcho.environ.parse_procfile(content).processes
                break

    if not processes:
        processes = get_app_processes(info)

    if process:
        if process not in processes:
            click.echo('Unknown process %s' % process)
            sys.exit(1)
        click.echo('Only running process %s' % process)
        processes = dict([(process, processes[process])])

    start_honcho(processes, {"HOST": host, "PORT": port})


@click.command('start', short_help='Start a single named process')
@click.argument('process')
@pass_script_info
def start_command(info, process):
    processes = get_app_processes(info)
    if process not in processes:
        click.echo('Unknown process %s' % process)
        sys.exit(1)
    processes = dict([(process, processes[process])])
    if info.debug:
        click.echo('Running %s' % processes[process])
    start_honcho(processes)


@click.command('echo-process-cmd', short_help='Echo the command line needed to run a process')
@click.argument('process')
@pass_script_info
def echo_process_command(info, process):
    processes = get_app_processes(info)
    if process not in processes:
        click.echo('Unknown process %s' % process)
        sys.exit(1)
    click.echo(processes[process])


# Unfortunately, there are no way to add options to run_simple
# without redifining the whole command

@click.command('serve', short_help='Runs a development server.')
@click.option('--host', '-h', default='127.0.0.1',
              help='The interface to bind to.')
@click.option('--port', '-p', default=5000,
              help='The port to bind to.')
@click.option('--reload/--no-reload', default=None,
              help='Enable or disable the reloader.  By default the reloader '
              'is active if debug is enabled.')
@click.option('--debugger/--no-debugger', default=None,
              help='Enable or disable the debugger.  By default the debugger '
              'is active if debug is enabled.')
@click.option('--eager-loading/--lazy-loader', default=None,
              help='Enable or disable eager loading.  By default eager '
              'loading is enabled if the reloader is disabled.')
@click.option('--with-threads/--without-threads', default=False,
              help='Enable or disable multithreading.')
@pass_script_info
def serve_command(info, host, port, reload, debugger, eager_loading,
                with_threads):
    """Runs a local development server for the Flask application.

    This local server is recommended for development purposes only but it
    can also be used for simple intranet deployments.  By default it will
    not support any sort of concurrency at all to simplify debugging.  This
    can be changed with the --with-threads option which will enable basic
    multithreading.

    The reloader and debugger are by default enabled if the debug flag of
    Flask is enabled and disabled otherwise.
    """
    from werkzeug.serving import run_simple
    debug = info.debug if info.debug is not None else info.env.lower() == "dev"
    if reload is None:
        reload = debug
    if debugger is None:
        debugger = debug
    if eager_loading is None:
        eager_loading = not reload

    app = DispatchingApp(info.load_app, use_eager_loading=eager_loading)

    # Extra startup messages.  This depends a but on Werkzeug internals to
    # not double execute when the reloader kicks in.
    if os.environ.get('WERKZEUG_RUN_MAIN') != 'true':
        # If we have an import path we can print it out now which can help
        # people understand what's being served.  If we do not have an
        # import path because the app was loaded through a callback then
        # we won't print anything.
        print(' * Using environment %s' % info.env)
        if info.app_import_path is not None:
            print(' * Serving Flask app "%s"' % info.app_import_path)
        if info.debug is not None:
            print(' * Forcing debug %s' % (info.debug and 'on' or 'off'))

    reloader_path = '.'
    if info.app_import_path:
        if os.path.isdir(info.app_import_path):
            reloader_path = info.app_import_path
        elif os.path.isfile(info.app_import_path):
            reloader_path = os.path.dirname(info.app_import_path)
    extra_files = get_reloader_extra_files(reloader_path)

    run_simple(host, port, app, use_reloader=reload, extra_files=extra_files,
               use_debugger=debugger, threaded=with_threads)


import functools
from flask import request, abort, current_app
from .utils import unknown_value


def _value_from_multidict(v):
    if len(v) == 0:
        return None
    elif len(v) == 1:
        return v[0]
    return v


class MissingRequestParam(Exception):
    pass


def get_request_param_value(name, location=None):
    if (not location or location == 'view_args') and name in request.view_args:
        return request.view_args[name]
    if not request.is_json:
        if not location or location == 'values':
            if name in request.values:
                return _value_from_multidict(request.values.getlist(name))
        elif location == 'args' and name in request.args:
            return _value_from_multidict(request.args.getlist(name))
        elif location == 'form' and name in request.form:
            return _value_from_multidict(request.form.getlist(name))
    if (not location or location == 'json') and request.is_json:
        data = request.get_json()
        if isinstance(data, dict) and name in data:
            return data[name]
    if (not location or location == 'files') and name in request.files:
        return _value_from_multidict(request.files.getlist(name))
    raise MissingRequestParam("Request parameter '%s' is missing" % name)


class RequestParam(object):
    def __init__(self, name, type=None, nullable=False, required=False, loader=None, validator=None,
                 dest=None, location=None, help=None, default=unknown_value, **loader_kwargs):
        self.name = name
        self.type = type
        self.nullable = nullable
        self.required = required
        self.loader = loader
        self.loader_kwargs = loader_kwargs
        self.validator = validator
        self.dest = dest
        self.location = location
        self.help = help
        self.default = default

    @property
    def names(self):
        return self.name if isinstance(self.name, tuple) else (self.name,)

    @property
    def dests(self):
        if not self.dest:
            return self.names
        elif not isinstance(self.dest, tuple):
            return (self.dest,)
        return self.dest

    def process(self):
        try:
            values = self.extract()
        except MissingRequestParam:
            if self.required and self.default is unknown_value:
                abort(400)
            elif not self.default is unknown_value:
                if not isinstance(self.name, tuple):
                    values = [self.default]
                else:
                    values = self.default
            else:
                return dict()

        values = self.load(*values)
        if not self.validate(*values):
            abort(400)
        if len(self.dests) != len(values):
            raise Exception('Mismatch between length of dest and number of values')
        return dict(zip(self.dests, values))

    def extract(self):
        values = []
        for name in self.names:
            value = get_request_param_value(name, self.location)
            if value is None and not self.nullable:
                raise MissingRequestParam("Request parameter '%s' is missing" % name)
            if self.type and value is not None:
                try:
                    if self.type is bool:
                        if isinstance(value, (str, unicode)):
                            value = value.lower() in ('true', '0')
                        else:
                            value = bool(value)
                    else:
                        value = self.type(value)
                except Exception as e:
                    current_app.logger.error(e)
                    abort(400)
            values.append(value)
        return values

    def load(self, *values):
        if self.loader:
            rv = self.loader(*values, **self.loader_kwargs)
            if not isinstance(rv, tuple):
                return (rv,)
            return rv
        return values

    def validate(self, *values):
        return not self.validator or self.validator(*values)


def request_param(name, cls=RequestParam, **kwargs):
    def decorator(func):
        if not hasattr(func, '_request_params'):
            func._request_params = []
        func._request_params.append(cls(name, **kwargs))
        return func
    return decorator


def partial_request_param(name=None, **kwargs):
    def decorator(name_=None, **kw):
        return request_param(name_ or name, **dict(kwargs, **kw))
    return decorator


def request_params(params):
    def decorator(func):
        for name, kwargs in params.iteritems():
            request_param(name, **kwargs)(func)
        return func
    return decorator


def request_param_loader(name, **kwargs):
    def decorator(loader_func):
        return request_param(name, loader=loader_func, **kwargs)
    return decorator


def partial_request_param_loader(name=None, **kwargs):
    def decorator(loader_func):
        def partial_gen(name_=None, **kw):
            return request_param(name_ or name, loader=loader_func, **dict(kwargs, **kw))
        return partial_gen
    return decorator


def process_request_params(params):
    values = {}
    for param in params:
        values.update(param.process())
    return values


def inject_request_params(func):
    @functools.wraps(func)
    def wrapper(**kwargs):
        if hasattr(func, '_request_params'):
            kwargs = process_request_params(func._request_params)
        return func(**kwargs)
    return wrapper


import click
from .signals import signal
from .utils import ClassBoundDecoratorMixin, get_reloader_extra_files
from .actions import ActionFunction
from .decorators import WithActionsDecorator
from flask import current_app
from flask.cli import pass_script_info as pass_script_info_decorator
from werkzeug.serving import run_with_reloader
import inspect
import contextlib
import subprocess


before_command_signal = signal("before_command")
after_command_signal = signal("after_command")


class CommandDecorator(ClassBoundDecoratorMixin):
    """Decorator to delay creation of click commands. Allows
    usage on instance methods.
    """
    def __init__(self, func, name=None, auto_params=True, echo=True, with_app_ctx=True, with_request_ctx=False,\
                 with_reloader=False, pass_script_info=False, cls=click.Command, **options):
        self.real_func = func
        if isinstance(self.real_func, ActionFunction):
            self.real_func = self.real_func.unbound_func
        if isinstance(self.real_func, WithActionsDecorator):
            self.real_func = self.real_func.unbound_func

        self.unbound_func = func
        self.func = self._wrap_func(func)
        self.name = name or self.real_func.__name__.replace("_", "-")
        self.echo = echo
        self.with_app_ctx = with_app_ctx
        self.with_request_ctx = with_request_ctx
        self.with_reloader = with_reloader
        self.pass_script_info = pass_script_info
        self.cls = cls
        self.options = options

        if auto_params:
            self.auto_create_params()

    def auto_create_params(self):
        argspec = inspect.getargspec(self.real_func)
        args = argspec.args
        if args[0] == "self":
            del args[0]
        if self.pass_script_info and args[0] == "info":
            del args[0]
        num_mandatory_args = len(args) - len(argspec.defaults) if argspec.defaults else len(args)
        if num_mandatory_args:
            mandatory_args = args[:num_mandatory_args]
            mandatory_args.reverse()
            for arg in mandatory_args:
                if not self.has_param(arg):
                    click.argument(arg)(self.unbound_func)
        if len(args) > num_mandatory_args:
            for i, arg in enumerate(args[num_mandatory_args:]):
                if not self.has_param(arg):
                    optname = "--%s" % arg.replace("_", "-")
                    default = argspec.defaults[i]
                    if isinstance(default, bool):
                        optname += "/--no-%s" % arg.replace("_", "-")
                    click.option(optname, default=default)(self.unbound_func)

    def has_param(self, name):
        try:
            for param in self.unbound_func.__click_params__:
                if param.name == name:
                    return True
        except AttributeError:
            pass
        return False

    def as_command(self):
        """Creates the click command wrapping the function
        """
        try:
            params = self.unbound_func.__click_params__
            params.reverse()
            del self.unbound_func.__click_params__
        except AttributeError:
            params = []
        help = inspect.getdoc(self.real_func)
        if isinstance(help, bytes):
            help = help.decode('utf-8')
        self.options.setdefault('help', help)
        @pass_script_info_decorator
        def callback(info, *args, **kwargs):
            if self.with_reloader:
                app = info.load_app()
                if app.debug:
                    def inner():
                        return self.command_callback(info, *args, **kwargs)
                    run_with_reloader(inner, extra_files=get_reloader_extra_files())
                    return
            self.command_callback(info, *args, **kwargs)
        return self.cls(name=self.name, callback=callback, params=params, **self.options)

    @contextlib.contextmanager
    def conditional_app_ctx(self, info):
        if self.with_app_ctx:
            with info.load_app().app_context():
                yield
        else:
            yield

    @contextlib.contextmanager
    def conditional_request_ctx(self):
        if self.with_request_ctx:
            kwargs = self.with_request_ctx
            if isinstance(kwargs, str):
                kwargs = {"path": kwargs}
            with current_app.test_request_context(**kwargs):
                yield
        else:
            yield

    def command_callback(self, info, *args, **kwargs):
        if self.pass_script_info:
            args = list(args)
            args.insert(0, info)
        before_command_signal.send(self)
        with self.conditional_app_ctx(info), self.conditional_request_ctx():
            rv = None
            try:
                rv = self.func(*args, **kwargs)
                if rv and self.echo:
                    click_echo_formated(rv)
            except Exception as e:
                if current_app.debug:
                    raise
                click.secho("ERROR of type %s: %s" % (type(e).__name__, e.message), fg="red", bold=True)
        after_command_signal.send(self)

    def format(self, value):
        if isinstance(value, (list, tuple, set)):
            return "\n".join([" - %s" % v for v in value])
        elif isinstance(value, dict):
            return "\n".join([" - %s: %s" % (k, v) for k, v in value.iteritems()])
        return value

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)


def click_echo_formated(value, **kwargs):
    if isinstance(value, (list, tuple, set)):
        value = "\n".join([" - %s" % v for v in value])
    elif isinstance(value, dict):
        value = "\n".join([" - %s: %s" % (k, v) for k, v in value.iteritems()])
    click.secho(str(value), **kwargs)


def command(*args, **kwargs):
    def decorator(f):
        return CommandDecorator(f, *args, **kwargs)
    return decorator

command.opt = click.option
command.arg = click.argument
command.echo = click_echo_formated


class ShellError(Exception):
    def __init__(self, returncode, stderr):
        super(ShellError, self).__init__(stderr)
        self.returncode = returncode
        self.stderr = stderr


def shell_exec(args, echo=True, fg="green", **kwargs):
    kwargs["stdout"] = subprocess.PIPE
    kwargs["stderr"] = subprocess.STDOUT
    p = subprocess.Popen(args, **kwargs)
    out, _ = p.communicate()
    if p.returncode > 0:
        raise ShellError(p.returncode, out)
    if echo:
        click.secho(out, fg=fg)
    return out
    


import click
import sys
import os
from flask.cli import FlaskGroup, ScriptInfo, shell_command
from ..declarative import create_app_from_cwd
from .run import run_command, start_command, echo_process_command, serve_command
from .scaffold import init_command, gen_command


class FrascoScriptInfo(ScriptInfo):
    @property
    def env(self):
        return os.environ.get("FRASCO_ENV", "prod")

    def load_app(self):
        if self.create_app is None and self.app_import_path is None:
            if os.path.exists("app.py"):
                self.app_import_path = "app"
            elif os.path.exists("app.yml"):
                self.create_app = lambda i: create_app_from_cwd()
        return super(FrascoScriptInfo, self).load_app()


def set_env_value(ctx, param, value):
    os.environ["FRASCO_ENV"] = value or "prod"


env_option = click.Option(['-e', '--env'],
    help='The runtime environment',
    callback=set_env_value, is_eager=True)


class FrascoGroup(FlaskGroup):
    def __init__(self, add_default_commands=True, add_env_option=True, **kwargs):
        params = list(kwargs.pop('params', None) or ())
        if add_env_option:
            params.append(env_option)
        super(FrascoGroup, self).__init__(add_default_commands=False, params=params, **kwargs)
        if add_default_commands:
            self.add_command(run_command)
            self.add_command(start_command)
            self.add_command(echo_process_command)
            self.add_command(serve_command)
            self.add_command(shell_command)
            self.add_command(init_command)
            self.add_command(gen_command)

    def main(self, *args, **kwargs):
        obj = kwargs.get('obj')
        if obj is None:
            obj = FrascoScriptInfo(create_app=self.create_app)
        kwargs['obj'] = obj
        kwargs.setdefault('auto_envvar_prefix', 'FRASCO')
        super(FrascoGroup, self).main(*args, **kwargs)


cli = FrascoGroup(help="""\
This shell command acts as general utility script for Frasco applications.

It will search for an app.yml file in the current directory or use the
one provided using the --app parameter or the FRASCO_APP environment variable.

The most useful command is "run".

Example usage:

  frasco --debug run
""")


def main(as_module=False):
    """This is copy/paste of flask.cli.main to instanciate our own group
    """
    this_module = __package__
    args = sys.argv[1:]

    if as_module:
        if sys.version_info >= (2, 7):
            name = 'python -m ' + this_module.rsplit('.', 1)[0]
        else:
            name = 'python -m ' + this_module

        # This module is always executed as "python -m flask.run" and as such
        # we need to ensure that we restore the actual command line so that
        # the reloader can properly operate.
        sys.argv = ['-m', this_module] + sys.argv[1:]
    else:
        name = None

    cli.main(args=args, prog_name=name)




import flask.signals
from .actions import load_grouped_actions, Context


_signals = flask.signals.Namespace()
signal = _signals.signal


def listens_to(name, sender=None, weak=True):
    """Listens to a named signal
    """
    def decorator(f):
        if sender:
            return signal(name).connect(f, sender=sender, weak=weak)
        return signal(name).connect(f, weak=weak)
    return decorator


def create_action_listener(event, actions):
    def callback(sender, **kwargs):
        with Context(dict(sender=sender, **kwargs)):
            actions.execute()
    signal(event).connect(callback, weak=False)


def create_listeners_from_config(config):
    for spec in config:
        create_action_listener(spec['name'], load_grouped_actions(spec))


from flask import render_template_string, url_for, json
from flask._compat import iteritems
from jinja2 import (BaseLoader, FileSystemLoader, ChoiceLoader, PrefixLoader,\
                    Environment as BaseEnvironment, nodes, TemplateNotFound)
from jinja2.ext import ExprStmtExtension
from ..utils import url_for_static, url_for_same, slugify
from jinja_macro_tags import MacroLoader, FileLoader, configure_environment as configure_macro_environment
from jinja_layout import LayoutExtension
import os
from .extensions import *
from helpers import *
from frasco import trans


class AliasLoader(BaseLoader):
    def __init__(self, aliases=None):
        self.aliases = aliases or {}

    def get_source(self, environment, template):
        if template in self.aliases:
            if isinstance(self.aliases[template], BaseLoader):
                return self.aliases[template].get_source(environment, template)
            return environment.loader.get_source(environment, self.aliases[template])
        raise TemplateNotFound(template)

    def list_templates(self):
        return self.aliases.keys()


class JinjaLoader(MacroLoader):
    """This Jinja loader replaces Flask's DispatchingJinjaLoader and adds
    multiple overriding points. Loaders can be added as a top level loader
    (first in the search path) using append() or to the lists feature_loaders
    and bottom_loaders. Loaders will be searched in the following order:
    top level loaders, the application's loader, loaders from blueprints
    added by the user, feature_loaders, loaders from blueprints provided
    by features, bottom_loaders.
    """
    def __init__(self, app):
        self.app = app
        self._loaders = []
        self.layout_loader = None
        self.feature_loaders = []
        self.macro_loaders = []
        self.bottom_loaders = []
        self.prefix_loader = PrefixLoader({
            # because we don't call MacroLoader's constructor to provide
            # our own loaders property, we define the prefix_loader with the
            # __macros__ entry
            "__macros__": ChoiceLoader(self.macro_loaders),
            # these prefixes allow for direct access to templates
            # defined by features
            "__super__": ChoiceLoader([
                ChoiceLoader(self.feature_loaders),
                ChoiceLoader(self._iter_feature_blueprint_loaders()),
                ChoiceLoader(self.bottom_loaders)])})

    def append(self, loader):
        self._loaders.append(loader)

    def add_path(self, path):
        self.append(FileSystemLoader(path))

    def add_prefix(self, prefix, loader):
        self.prefix_loader.mapping[prefix] = loader

    def add_super(self, loader):
        self.prefix_loader.mapping["__super__"].loaders.insert(0, loader)

    def set_layout_alias(self, template):
        if not self.layout_loader:
            self.layout_loader = AliasLoader()
        self.layout_loader.aliases["layout.html"] = template

    @property
    def loaders(self):
        # loaders added directly on this loader
        for loader in self._loaders:
            yield loader

        # app loader
        loader = self.app.jinja_loader
        if loader is not None:
            yield loader

        # loaders for blueprints added by the user
        for loader in self._iter_app_blueprint_loaders():
            yield loader

        if self.layout_loader:
            yield self.layout_loader

        # loaders that overrides blueprints provided by features
        for loader in self.feature_loaders:
            yield loader

        # loaders for blueprints provided by features
        for loader in self._iter_feature_blueprint_loaders():
            yield loader

        # loaders that are at the bottom
        for loader in self.bottom_loaders:
            yield loader

        yield self.prefix_loader

    def _iter_app_blueprint_loaders(self):
        for name, blueprint in iteritems(self.app.blueprints):
            if name in self.app.feature_blueprint_names:
                continue
            loader = blueprint.jinja_loader
            if loader is not None:
                yield loader

    def _iter_feature_blueprint_loaders(self):
        for name in self.app.feature_blueprint_names:
            loader = self.app.blueprints[name].jinja_loader
            if loader is not None:
                yield loader


class Environment(BaseEnvironment):
    def __init__(self, app, **options):
        if 'autoescape' not in options:
            options['autoescape'] = app.select_jinja_autoescape
        if 'auto_reload' not in options:
            options['auto_reload'] = app.debug \
                or app.config['TEMPLATES_AUTO_RELOAD']
        super(Environment, self).__init__(**options)
        self.app = app
        configure_environment(self)
        self.globals["config"] = app.config


def configure_environment(env, with_blocks=True, with_macros=True, with_layout=True, with_trans=True):
    env.add_extension(ExprStmtExtension)
    env.add_extension(RemoveYamlFrontMatterExtension)
    configure_macro_environment(env, wrap_loader=False)
    env.add_extension(LayoutExtension)
    env.add_extension(FlashMessagesExtension)
    env.globals.update(real_dict=dict,
                       getattr=getattr,
                       url_for=url_for,
                       url_for_static=url_for_static,
                       url_for_same=url_for_same,
                       html_tag=html_tag,
                       html_attributes=html_attributes,
                       plural=plural)
    env.filters.update(slugify=slugify,
                       tojson=json.tojson_filter,
                       nl2br=nl2br,
                       timeago=timeago)

    if with_trans:
        env.globals.update(translate=trans.translate,
                           ntranslate=trans.ntranslate,
                           _=trans.translate)
        env.filters.update(datetimeformat=trans.format_datetime,
                           dateformat=trans.format_date,
                           timeformat=trans.format_time)

    if with_blocks:
        env.add_extension(jinja_block_as_fragment_extension("content"))
    if with_macros:
        env.macros.register_file(os.path.join(os.path.dirname(__file__), "macros.html"), "frasco.html")
    if with_layout:
        env.loader.bottom_loaders.append(FileLoader(
            os.path.join(os.path.dirname(__file__), "layout.html"), "frasco_layout.html"))
        env.loader.set_layout_alias("frasco_layout.html")
    env.default_layout = "layout.html"


def render_layout(layout_name, content, **context):
    """Uses a jinja template to wrap the content inside a layout.
    Wraps the content inside a block and adds the extend statement before rendering it
    with jinja. The block name can be specified in the layout_name after the filename separated
    by a colon. The default block name is "content".
    """
    layout_block = "content"
    if ":" in layout_name:
        layout_name, layout_block = layout_name.split(":")
    tpl = '{%% extends "%s" %%}{%% block %s %%}%s{%% endblock %%}' % (layout_name, layout_block, content)
    return render_template_string(tpl, **context)


def get_template_source(app, filename):
    (source, _, __) = app.jinja_env.loader.get_source(app.jinja_env, filename)
    return source


def parse_template(app, filename):
    """Parses the given template using the jinja environment of the given app
    and returns the AST. ASTs are cached in parse_template.cache
    """
    if not hasattr(parse_template, "cache"):
        parse_template.cache = {}
    if filename not in parse_template.cache:
        source = get_template_source(app, filename)
        parse_template.cache[filename] = app.jinja_env.parse(source, filename=filename)
    return parse_template.cache[filename]


def jinja_node_to_python(node):
    """Converts a Jinja2 node to its python equivalent
    """
    if isinstance(node, nodes.Const):
        return node.value
    if isinstance(node, nodes.Neg):
        return -jinja_node_to_python(node.node)
    if isinstance(node, nodes.Name):
        return node.name
    if isinstance(node, (nodes.List, nodes.Tuple)):
        value = []
        for i in node.items:
            value.append(jinja_node_to_python(i))
        return value
    if isinstance(node, nodes.Dict):
        value = {}
        for pair in node.items:
            value[pair.key.value] = jinja_node_to_python(pair.value)
        return value
    raise Exception("Cannot convert jinja nodes to python")


from .actions import (ActionFunction, ActionList, current_context, ContextExitException,\
                     ensure_context)
from .decorators import WithActionsDecorator, register_hooks, preprend_base_url_to_expose
from .views import Blueprint, as_view, ActionsView, full_exec_request_actions
from .request import process_request_params
from .utils import logger
from flask import Response, json, request, current_app, jsonify, abort, make_response
from werkzeug.local import LocalProxy
from werkzeug.exceptions import HTTPException
from werkzeug.wrappers import Response as WResponse
import inspect
import functools


class ServiceActionsView(ActionsView):
    """View class for service views created using the ServiceLoader
    """
    def execute_actions(self, *args, **kwargs):
        func = None
        if self.func:
            func = lambda: self.func(*args, **kwargs)
        def render():
            # the render func is only called if the actions do not
            # exit the context, thus we return an empty json response
            return jsonify()
        rv = full_exec_request_actions(self.actions, func=func, render_func=render)
        # child actions can use "return" to return a value which
        # can be something else than a proper Response instance.
        # In this case we encode the return value to json
        if isinstance(rv, (Response, WResponse)):
            return rv
        return json.dumps(rv), {'Content-Type': 'application/json;charset=UTF-8'}


def patch_action(action):
    """This will add raising TriggerActionGroupException when api_success
    or api_error have been used. This allows nice logic flows when using
    a service as an action
    """
    old_execute = action.execute
    def execute(self):
        with ensure_context():
            try:
                if old_execute(self):
                    current_context.trigger_action_group(action.name + "_success")
            except ServiceError as e:
                current_context['service_error'] = e
                current_context.trigger_action_group(action.name + "_error")
    action.execute = execute
    return action


def marshall_with(marshaller):
    def decorator(func):
        func._marshall_with = marshaller
        return func
    return decorator


def marshall_many_with(marshaller):
    def many_marshaller(data, func_kwargs):
        return [marshaller(i, func_kwargs) for i in data]
    def decorator(func):
        func._marshall_with = many_marshaller
        return func
    return decorator


def wrap_view_func(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            if hasattr(func, '_request_params'):
                kwargs = process_request_params(func._request_params)
            rv = func(**kwargs)
            if hasattr(func, '_marshall_with'):
                return func._marshall_with(rv, kwargs)
            return rv
        except ServiceError as e:
            return make_response(json.dumps({"error": e.message}), e.http_code)
    if hasattr(func, '_request_params'):
        wrapper._request_params = func._request_params
    return wrapper


def get_current_app_service(name):
    return current_app.services[name]


_services_proxies = {}

def service_proxy(name):
    if name not in _services_proxies:
        _services_proxies[name] = LocalProxy(lambda: get_current_app_service(name))
    return _services_proxies[name]


def pass_service(*names):
    """Injects a service instance into the kwargs
    """
    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            for name in names:
                kwargs[name] = service_proxy(name)
            return f(*args, **kwargs)
        return wrapper
    return decorator


class ServiceError(Exception):
    def __init__(self, message, http_code=500):
        self.message = message
        self.http_code = http_code


class ServiceMeta(type):
    def __init__(cls, name, bases, attrs):
        type.__init__(cls, name, bases, attrs)
        cls.current = service_proxy(attrs.get('name') or name)


class Service(object):
    name = None
    view_class = ServiceActionsView
    blueprint_class = Blueprint
    url_prefix = None
    __metaclass__ = ServiceMeta

    def __init__(self):
        self.name = self.name or self.__class__.__name__
        self.actions = ActionList()
        self.views = []
        self.hooks = []
        for attr in dir(self):
            attr = getattr(self, attr)
            if isinstance(attr, ActionFunction):
                self.actions.append(patch_action(attr.action))
                if hasattr(attr, 'urls'):
                    preprend_base_url_to_expose(self.url_prefix, attr)
                    func = attr.unbound_func if isinstance(attr.unbound_func, WithActionsDecorator) else attr.func
                    view_func = wrap_view_func(func)
                    self.views.append(as_view(name=attr.unbound_func.__name__, url_rules=attr.urls, view_class=self.view_class)(view_func))
            elif hasattr(attr, 'urls'):
                preprend_base_url_to_expose(self.url_prefix, attr)
                view_func = wrap_view_func(attr)
                self.views.append(as_view(name=attr.__name__, url_rules=attr.urls, view_class=self.view_class)(view_func))
            elif hasattr(attr, 'hooks'):
                self.hooks.append(attr)

    @property
    def exposed(self):
        return len(self.views) > 0 or len(self.hooks) > 0

    @property
    def blueprint_name(self):
        return "%s_service" % self.name

    def as_blueprint(self):
        bp = self.blueprint_class(self.blueprint_name, self.__class__.__module__)
        for view in self.views:
            bp.add_view(view)
        for attr in self.hooks:
            register_hooks(attr, attr.hooks, bp)
        return bp


from jinja2 import nodes
from jinja2.ext import Extension
from flask import get_flashed_messages
from ..utils import remove_yaml_frontmatter


__all__ = ('jinja_fragment_extension', 'jinja_block_as_fragment_extension',
           'RemoveYamlFrontMatterExtension', 'FlashMessagesExtension')


def parse_block_signature(parser):
    args = []
    kwargs = []
    require_comma = False

    while parser.stream.current.type != 'block_end':
        if require_comma:
            parser.stream.expect('comma')
            # support for trailing comma
            if parser.stream.current.type == 'block_end':
                break
        if parser.stream.current.type == 'name' and \
           parser.stream.look().type == 'assign':
            key = parser.stream.current.value
            parser.stream.skip(2)
            value = parser.parse_expression()
            kwargs.append(nodes.Keyword(key, value,
                                        lineno=value.lineno))
        else:
            args.append(parser.parse_expression())
        require_comma = True

    return args, kwargs


def jinja_fragment_extension(tag, endtag=None, name=None, tag_only=False, allow_args=True, callblock_args=None):
    """Decorator to easily create a jinja extension which acts as a fragment.
    """
    if endtag is None:
        endtag = "end" + tag

    def decorator(f):
        def parse(self, parser):
            lineno = parser.stream.next().lineno
            args = []
            kwargs = []
            if allow_args:
                args, kwargs = parse_block_signature(parser)

            call = self.call_method("support_method", args, kwargs, lineno=lineno)
            if tag_only:
                return nodes.Output([call], lineno=lineno)

            call_args = []
            if callblock_args is not None:
                for arg in callblock_args:
                    call_args.append(nodes.Name(arg, 'store', lineno=lineno))

            body = parser.parse_statements(['name:' + endtag], drop_needle=True)
            return nodes.CallBlock(call, call_args, [], body, lineno=lineno)

        def support_method(self, *args, **kwargs):
            return f(*args, **kwargs)

        attrs = {"tags": set([tag]), "parse": parse, "support_method": support_method}
        return type(name or f.__name__, (Extension,), attrs)

    return decorator


class BaseJinjaBlockAsFragmentExtension(Extension):
    def parse(self, parser):
        lineno = parser.stream.next().lineno
        body = parser.parse_statements(['name:' + self.end_tag], drop_needle=True)
        return nodes.Block(self.block_name, body, False, lineno=lineno)


def jinja_block_as_fragment_extension(name, tagname=None, classname=None):
    """Creates a fragment extension which will just act as a replacement of the
    block statement.
    """
    if tagname is None:
        tagname = name
    if classname is None:
        classname = "%sBlockFragmentExtension" % name.capitalize()
    return type(classname, (BaseJinjaBlockAsFragmentExtension,), {
        "tags": set([tagname]), "end_tag": "end" + tagname, "block_name": name})


class RemoveYamlFrontMatterExtension(Extension):
    """Jinja extension to remove YAML front-matters
    """
    def preprocess(self, source, name, filename=None):
        before = ""
        if source.startswith("{% load_macro"):
            before, source = source.split("\n", 1)
        return before + remove_yaml_frontmatter(source)


@jinja_fragment_extension("flash_messages", endtag="endflash", allow_args=False,
                          callblock_args=["message", "category"])
def FlashMessagesExtension(caller=None):
    messages = get_flashed_messages(with_categories=True)
    html = ""
    if messages:
        for cat, msg in messages:
            html += caller(msg, cat)
    return html



from .app import DeclarativeFrasco, DeclarativeFrascoFactory, create_app_from_cwd, decl_app
from .views import DeclarativeBlueprint, DeclarativeView
from .loaders import FileLoader, ViewsLoader, ActionsLoader, FeaturesLoader


import os
import sys
import yaml
from ..app import Frasco
from ..actions import load_actions, load_grouped_actions, Context
from ..utils import import_string, logger, deep_update_dict
from ..signals import create_listeners_from_config
from ..views import Blueprint
from .loaders import FeaturesLoader, ActionsLoader, ViewsLoader, ServicesLoader
from flask.helpers import get_root_path
from flask import render_template
from jinja2 import FileSystemLoader
from werkzeug.local import LocalProxy


_decl_app = None

def get_current_decl_app():
    return _decl_app

decl_app = LocalProxy(get_current_decl_app)


def create_app_from_cwd(import_name="__main__", **kwargs):
    return DeclarativeFrasco(import_name, root_path=os.getcwd(),
        set_as_current=True, **kwargs)


class DeclarativeFrasco(Frasco):
    def __new__(cls, import_name, **kwargs):
        decl_kwargs = {k: kwargs.pop(k) for k in kwargs.keys() if k in
                        ('config_filename', 'view_folder', 'feature_folder',
                         'action_folder')}
        factory = DeclarativeFrascoFactory(**decl_kwargs)
        return factory.create_app(import_name, **kwargs)


class DeclarativeFrascoFactory(object):
    """Creates an app from the given configuration file. Will load features, actions
    and blueprints relative to the configuration file
    """

    def __init__(self, config_filename="app.yml", view_folder="views",
                 feature_folder="features", action_folder="actions"):
        self.config_filename = config_filename
        self.view_folder = view_folder
        self.feature_folder = feature_folder
        self.action_folder = action_folder

    def create_app(self, import_name, set_as_current=False, **kwargs):
        app = Frasco(import_name, **kwargs)
        if set_as_current:
            global _decl_app
            _decl_app = app
        self.setup_app(app)
        return app

    def setup_app(self, app):
        self.configure_app(app, False)
        self.load_actions((app))
        self.load_services(app)
        self.load_blueprints(app)

    def configure_app(self, app, with_feature_blueprints=True, env=None):
        config_path = os.path.join(app.root_path, self.config_filename)
        view_path = os.path.join(app.root_path, self.view_folder)
        feature_path = os.path.join(app.root_path, self.feature_folder)

        app.jinja_loader.loaders.insert(0, FileSystemLoader(view_path))
        app.view_folder = self.view_folder
        app.config.from_yaml(config_path, silent=True)

        env = env or os.environ.get("FRASCO_ENV") or "prod"
        filename, ext = os.path.splitext(config_path)
        env_filename = filename + "-" + env + ext
        app.config['ENV'] = env
        app.config.from_yaml(env_filename, silent=True, deep_update=True)

        if "INCLUDE_FILES" in app.config:
            for spec in app.config["INCLUDE_FILES"]:
                if not isinstance(spec, dict):
                    spec = {"filename": spec}
                filename = os.path.join(os.path.dirname(config_path), spec.pop("filename"))
                app.config.from_yaml(filename, **spec)

        if os.path.exists(feature_path):
            sys.path.insert(0, feature_path)
        for feature in FeaturesLoader().load(app.config.get('FEATURES', [])):
            app.register_feature(feature, with_blueprints=with_feature_blueprints)
            feature.init_declarative(app)

        if "ERROR_HANDLERS" in app.config:
            for code, template in app.config["ERROR_HANDLERS"].iteritems():
                self.register_error_handler(app, code, template)

        if "BEFORE_REQUEST_ACTIONS" in app.config:
            app.request_actions.extend(load_actions(app.config["BEFORE_REQUEST_ACTIONS"], 'before'))
        if "AFTER_REQUEST_ACTIONS" in app.config:
            app.request_actions.extend(load_actions(app.config["AFTER_REQUEST_ACTIONS"], 'after'))

        create_listeners_from_config(app.config.get("SIGNAL_LISTENERS", []))

        for fqdn in app.config.get("IMPORTS", []):
            import_string(fqdn)

    def register_error_handler(self, app, code, template):
        @app.errorhandler(code)
        def func(error):
            return render_template(template, error=error), code

    def load_actions(self, app):
        action_path = os.path.join(app.root_path, self.action_folder)
        action_package = self.action_folder.replace("/", ".")
        if os.path.exists(action_path):
            loader = ActionsLoader(action_path, pypath=action_package)
            for action in loader.load(app):
                app.actions.register(action)

    def load_services(self, app):
        loader = ServicesLoader(os.path.join(app.root_path, "services"))
        for service in loader.load(app):
            app.register_service(service)

    def load_blueprints(self, app):
        view_path = os.path.join(app.root_path, self.view_folder)
        view_package = self.view_folder.replace("/", ".")
        if os.path.exists(view_path):
            loader = ViewsLoader(view_path, pypath=view_package)
            for obj in loader.load(app):
                if isinstance(obj, Blueprint):
                    app.register_blueprint(obj)
                else:
                    app.add_view(obj)

        for feature in app.features:
            feature.init_blueprints(app)



from .context import (current_context, ensure_context, ContextExitException,\
                      TriggerActionGroupException, Context, request_context)

from .core import (OptionMissingError, InvalidOptionError, Action, ActionFunction,\
                   ActionFunction, ActionProxy)

from .decorators import (action_decorator, WithItemsDecorator, WhenDecorator,\
                         IgnoreErrorDecorator, OnlyOnceDecorator)

from .registry import (action_resolver, ActionNotFoundError, ActionRegistry,\
                       DelayedInstanciationAction, resolve_action)

from .helpers import (ActionList, InvalidActionError, execute_action, execute_actions,\
                      ReturnValueException, ActionsAction)

from .loaders import load_actions, load_grouped_actions, create_action_from_dict

from .common import common_actions


import click
from flask import current_app
from flask.cli import with_appcontext
from .run import procfile_processes
import os
import random


def gen_secret_key():
    return ''.join([random.SystemRandom().choice('abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*(-_=+)') for i in range(50)])


@click.command("init")
@click.argument('path', default='.')
def init_command(path):
    if not os.path.exists(path):
        os.makedirs(path)
    if not os.path.exists(os.path.join(path, 'views')):
        os.mkdir(os.path.join(path, 'views'))
    secret_key = gen_secret_key()
    with open(os.path.join(path, 'app.yml'), 'w') as f:
        f.write("secret_key: \"%s\"\n" % secret_key)
    with open(os.path.join(path, 'app-dev.yml'), 'w') as f:
        f.write("debug: true\n")
    with open(os.path.join(path, 'views/index.html'), 'w') as f:
        f.write("---\nurl: /\n---\n{% use_layout %}\nHello world from Frasco")


@click.group("gen")
def gen_command():
    """Helpers to generate useful files"""


@gen_command.command('secret-key')
def gen_secret_key_cmd():
    click.echo(gen_secret_key())


@gen_command.command("app.py")
def gen_apppy():
    """Generates an app.py file
    """
    with open("app.py", "w") as f:
        f.write(("from frasco import DeclarativeFrasco\n\n\n"
                 "app = DeclarativeFrasco(__name__)"))


@gen_command.command("procfile")
@click.option("--wsgi", default=None,
    help="Name of the file containing the application")
@click.option("--dev/--no-dev", default=False,
    help="Whether to generate a Procfile for the dev environment")
@with_appcontext
@click.pass_context
def gen_procfile(ctx, wsgi, dev):
    """Generates Procfiles which can be used with honcho or foreman.
    """
    if wsgi is None:
        if os.path.exists("wsgi.py"):
            wsgi = "wsgi.py"
        elif os.path.exists("app.py"):
            wsgi = "app.py"
        else:
            wsgi = "app.py"
            ctx.invoke(gen_apppy)

    def write_procfile(filename, server_process, debug):
        processes = [server_process] + current_app.processes
        procfile = []
        for name, cmd in procfile_processes(processes, debug).iteritems():
            procfile.append("%s: %s" % (name, cmd))
        with open(filename, "w") as f:
            f.write("\n".join(procfile))

    write_procfile("Procfile", ("web", ["gunicorn", wsgi]), False)
    if dev:
        write_procfile("Procfile.dev", ("web", ["frasco", "serve"]), True)


from .actions import ActionRegistry, ActionFunction
from .commands import CommandDecorator
from .utils import RequirementMissingError, deep_update_dict, import_string, find_classes_in_module
from werkzeug.local import LocalProxy
from flask import current_app
import click
import inspect
import functools
from collections import OrderedDict


class FeatureError(Exception):
    pass


class Feature(object):
    """A feature configures an app and bundles actions.
    Can also register blueprints.
    """
    name = None
    # Name of other features which this one requires
    requires = None
    # Default options
    defaults = None
    # A list of Action classes
    actions = None
    # A list of blueprint instances that will be registered on the app
    blueprints = None
    # A list of click commands that will be registered on the app
    commands = None
    # Group commands declared using the @command() on members of this class
    # under a command group. This value can be the group name which would be
    # replaced with an instance of click.Group at instanciation or a click.Group object.
    # If False, no group will be used and commands will be added to the commands
    # attribute
    command_group = None
    # A list of hooks
    hooks = None
    # ignore some attributes when looking for annotations
    ignore_attributes = None

    def __init__(self, options=None, **kwargs):
        if not self.name:
            raise FeatureError("A name property must be defined on Feature classes")
        self.options = dict(self.defaults or {})
        if options is not None:
            self.options.update(options)
        self.options.update(kwargs)
        self.actions = ActionRegistry(self.actions)

        # We only do attribute discovery at instanciation time
        # because we want decorators used on methods to be bound
        # on the instance
        commands = []
        for name in dir(self):
            if name == "__class__" or (self.ignore_attributes is not None and name in self.ignore_attributes):
                continue
            attr = getattr(self, name)
            # Adds function decorated with @command to commands
            if isinstance(attr, CommandDecorator):
                commands.append(attr.as_command())
                attr = attr.unbound_func
            # Adds function decorated with @action to actions
            if isinstance(attr, ActionFunction):
                self.actions.register(attr)
            # Adds function decorator with @hook to hooks
            elif hasattr(attr, 'hooks'):
                if self.hooks is None:
                    self.hooks = []
                self.hooks.append((attr, attr.hooks))

        if commands:
            if not self.commands:
                self.commands = []
            if self.command_group is None or isinstance(self.command_group, str):
                name = self.command_group or self.name
                self.command_group = click.Group(name=name, help=inspect.getdoc(self))
            if self.command_group:
                for cmd in commands:
                    self.command_group.add_command(cmd)
                self.commands.append(self.command_group)
            else:
                self.commands.extend(commands)

    def init_app(self, app):
        pass

    def init_declarative(self, app):
        pass

    def init_blueprints(self, app):
        if self.blueprints:
            for bp in self.blueprints:
                kwargs = {}
                if isinstance(bp, tuple):
                    bp, kwargs = bp
                if inspect.isclass(bp):
                    bp = bp()
                elif callable(bp):
                    bp = bp(app)
                app.register_blueprint(bp, **kwargs)
                app.feature_blueprint_names.add(bp.name)


class FeaturesLoader(object):
    """Loads Feature classes from the FEATURES config key of the app
    """
    def load(self, config):
        merged_config = OrderedDict()
        for featurespec in config:
            name = featurespec
            options = {}
            if isinstance(featurespec, Feature):
                merged_config[featurespec.name] = featurespec
                continue
            if isinstance(featurespec, tuple):
                name, options = featurespec
            elif isinstance(featurespec, dict):
                name, options = featurespec.popitem()
            name = name.replace("-", "_")
            if name in merged_config:
                deep_update_dict(merged_config[name], options)
            else:
                merged_config[name] = options

        features = []
        for name, options in merged_config.items():
            if isinstance(options, Feature):
                features.append(options)
            else:
                features.append(self.load_feature(name)(options))

        return features

    def load_feature(self, name):
        if inspect.isclass(name) and issubclass(name, Feature):
            return name
        
        features = None
        if "." not in name:
            features = self.try_import("frasco_" + name)
        if features is None:
            features = self.try_import(name)

        if not features:
            raise RequirementMissingError("Cannot find a feature named '%s'" % name)
        if len(features) > 1:
            raise RequirementMissingError("The feature name '%s' references a module with than one Feature class" % name)
        return features[0]

    def try_import(self, name):
        try:
            imported = import_string(name)
        except ImportError:
            return None
        if inspect.ismodule(imported):
            # Gives the possibility to reference a module and auto-discover the Feature class
            return find_classes_in_module(imported, (Feature,))
        return [imported]


class FeatureRegistry(object):
    def __init__(self):
        self.features = {}

    def register(self, feature):
        self.features[feature.name] = feature

    def names(self):
        return self.features.keys()

    def exists(self, name):
        return name in self.features

    def __getitem__(self, name):
        if name not in self.features:
            raise RequirementMissingError("Missing feature '%s'" % name)
        return self.features[name]

    def __getattr__(self, name):
        return self[name]

    def __contains__(self, name):
        return name in self.features

    def __iter__(self):
        return self.features.itervalues()

    def __len__(self):
        return len(self.features)


def get_current_app_feature(name):
    return current_app.features[name]


_feature_proxies = {}

def feature_proxy(name):
    if name not in _feature_proxies:
        _feature_proxies[name] = LocalProxy(lambda: get_current_app_feature(name))
    return _feature_proxies[name]


def pass_feature(*feature_names):
    """Injects a feature instance into the kwargs
    """
    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            for name in feature_names:
                kwargs[name] = feature_proxy(name)
            return f(*args, **kwargs)
        return wrapper
    return decorator